/*
 *  segment.rs
 *
 *  Copyright (C) 2020 Qarva
 *  Sun Dro <skalatozishvili@qarva.com>
 *
 * Collect QMPP packet fragments and 
 * convert segments into TS payload
 */

extern crate bytebuffer;
use bytebuffer::ByteBuffer;

use crate::qws::packet::QPacket;
use std::vec::Vec;

const RTP_HEADER_SIZE: usize = 28;
const QWS_FRAME_SIZE: usize = 1316;

pub struct QSegment {
    payload: ByteBuffer,
    pub size: usize,
    pub pts: u64,
    pub id: u32
}

impl QSegment {
    pub fn new() -> QSegment {
        QSegment {
            payload: ByteBuffer::new(),
            size: 0,
            pts: 0,
            id: 0
        }
    }

    pub fn copy(&mut self) -> QSegment {
        QSegment {
            payload: ByteBuffer::from_bytes(&self.get_payload()),
            size: self.size,
            pts: self.pts,
            id: self.id
        }
    }

    pub fn to_ts_bytes(&mut self) -> Vec<u8> {
        let payload = self.get_payload();
        let mut buffer = ByteBuffer::new();
        let mut offset: usize = 0;

        while offset < self.size as usize {
            offset += RTP_HEADER_SIZE;
            let end = offset + QWS_FRAME_SIZE;
            let slice = &payload[offset..end];
            buffer.write_bytes(slice);
            offset += QWS_FRAME_SIZE;
        }

        buffer.to_bytes()
    }

    pub fn get_payload(&mut self) -> Vec<u8> {
        self.payload.to_bytes()
    }
}

pub struct QSegmenter {
    segments: Vec<QSegment>,
    packets: Vec<QPacket>,
    fragments_max: usize
}

impl QSegmenter {
    pub fn new(frags: usize) -> QSegmenter {
        QSegmenter {
            segments: Vec::new(),
            packets: Vec::new(),
            fragments_max: frags
        }
    }

    pub fn push_packet(&mut self, packet: &mut QPacket) -> bool {
        self.packets.push(packet.copy());
        let segment = packet.header.segment_id;
        let count = self.count_packets(segment);

        if count >= self.fragments_max as usize {
            let mut buffer = ByteBuffer::new();
            let last = self.fragments_max + 1;

            for frag in 1..last as u16 {
                let mut pkt = self.pop_packet(segment, frag).unwrap();
                buffer.write_bytes(&pkt.get_payload());
            }

            if buffer.len() == packet.header.segment_size as usize {
                self.segments.push(QSegment {
                    payload: buffer,
                    size: packet.header.segment_size,
                    pts: packet.header.pts,
                    id: packet.header.segment_id
                });

                println!("p({})[*] x Segmented TS: id({}), size({})", 
                    packet.header.fragment_id,
                    packet.header.segment_id,
                    packet.header.segment_size);
                return true;
            }
        }

        false
    }

    pub fn pop_segment(&mut self) -> Option<QSegment> {
        if self.segments.is_empty() { return None; }
        Some(self.segments.remove(0))
    }

    pub fn pop_packet(&mut self, id: u32, frag: u16) -> Option<QPacket> {
        for index in 0..self.packets.len() {
            if self.packets[index].header.segment_id == id && 
                self.packets[index].header.fragment_id == frag {
                    return Some(self.packets.remove(index));
            }
        }
        None
    }

    pub fn count_packets(&mut self, id: u32) -> usize {
        let mut packet_count: usize = 0;
        for packet in &mut self.packets {
            let seg = packet.header.segment_id;
            if seg == id { packet_count += 1; }
        }
        packet_count
    }
}